(*
#$Id: biject.sml 10 2011-04-13 08:39:38Z ns441 $
#
#Part of the "prefactor" experiment. Translates between datatypes targeted by parser and Isabelle-generated code
#
#Nik Sultana
#www.cl.cam.ac.uk/~ns441
#
*)

structure Biject : sig
  val isabellise_type : PCFP1_Syntax.typ -> CR.tipi
  val isabellise_term : string PCFP1_Syntax.trm -> CR.lExp
  val deisabellise_type : CR.tipi -> PCFP1_Syntax.typ
  val deisabellise_term : CR.lExp -> string PCFP1_Syntax.trm
end = struct

open PCFP1_Syntax

fun isabellise_type (*PCFP1_Syntax.typ -> CR.tipi*)
    (Type (ty,fill)) =
    case ty of
	T_Nat           => CR.Base (CR.Nt, fill)
      | T_Fun (ty1,ty2) => CR.Funk ((isabellise_type ty1), (isabellise_type ty2), fill)
      | T_Unit          => CR.Unit fill
      | T_Sum (ty1,ty2) => CR.Sumtp ((isabellise_type ty1), (isabellise_type ty2), fill)

fun isabellise_term (*PCFP1_Syntax.trm -> CR.lExp*)
    (Term (t,fill)) =
    case t of
	E_Var v                  => CR.LVar (v, fill)
      | E_Lam (v,ty,t')          => CR.LLam (v,isabellise_type ty,isabellise_term t', fill)
      | E_App (t1,t2)            => CR.LApp (isabellise_term t1, isabellise_term t2, fill)
      | E_Un                     => CR.Unity fill
      | E_Fix (v,ty,t')          => CR.Fp (v,isabellise_type ty,isabellise_term t', fill)
      | E_InL (ty,t')            => CR.InL (isabellise_type ty,isabellise_term t', fill)
      | E_InR (ty,t')            => CR.InR (isabellise_type ty,isabellise_term t', fill)
      | E_Case (tL,vL,t',vR,tR)  => CR.Caseof (isabellise_term tL,vL,isabellise_term t',vR,isabellise_term tR, fill)
      | E_Zero                      => CR.Zero fill
      | E_Succ t'                   => CR.Succ (isabellise_term t', fill)
      | E_Pred t'                   => CR.Pred (isabellise_term t', fill)
      | E_IfZero (t1,t2,t3)         => CR.Ifz (isabellise_term t1, isabellise_term t2, isabellise_term t3, fill)

fun deisabellise_type (*CR.tipi -> PCFP1_Syntax.typ*)
    ty =
    case ty of
	CR.Base (CR.Nt, fill)     => Type (T_Nat, fill)
      | CR.Funk (ty1, ty2, fill)  => Type (T_Fun (deisabellise_type ty1,deisabellise_type ty2), fill)
      | CR.Unit fill              => Type (T_Unit, fill)
      | CR.Sumtp (ty1, ty2, fill) => Type (T_Sum (deisabellise_type ty1,deisabellise_type ty2), fill)

fun deisabellise_term (*CR.lExp -> PCFP1_Syntax.trm*)
    t =
    case t of
	CR.LVar (v, fill)               => Term (E_Var v, fill)
      | CR.LLam (v,ty,t',fill)          => Term (E_Lam (v,deisabellise_type ty,deisabellise_term t'), fill)
      | CR.LApp (t1,t2,fill)            => Term (E_App (deisabellise_term t1,deisabellise_term t2), fill)
      | CR.Unity fill                   => Term (E_Un, fill)
      | CR.Fp (v,ty,t',fill)            => Term (E_Fix (v,deisabellise_type ty,deisabellise_term t'), fill)
      | CR.InL (ty,t',fill)             => Term (E_InL (deisabellise_type ty,deisabellise_term t'), fill)
      | CR.InR (ty,t',fill)             => Term (E_InR (deisabellise_type ty,deisabellise_term t'), fill)
      | CR.Caseof (tL,vL,t',vR,tR,fill) => Term (E_Case (deisabellise_term tL,vL,deisabellise_term t',vR,deisabellise_term tR), fill)
      | CR.Zero fill                    => Term (E_Zero, fill)
      | CR.Succ (t', fill)              => Term (E_Succ (deisabellise_term t'), fill)
      | CR.Pred (t', fill)              => Term (E_Pred (deisabellise_term t'), fill)
      | CR.Ifz (t1,t2,t3, fill)         => Term (E_IfZero (deisabellise_term t1,deisabellise_term t2,deisabellise_term t3), fill)
end
