(*
#$Id: isabelle.sml 32 2011-07-20 23:47:15Z ns441 $
#
#Part of the "prefactor" experiment. This file was initially generated by Isabelle, then edited slightly to integrate it with the rest of the code and add debugging info.
#
#Nik Sultana
#www.cl.cam.ac.uk/~ns441
#
*)

structure HOL : sig
  type 'a eq
  val eq : 'a eq -> 'a -> 'a -> bool
  val eqa : 'a eq -> 'a -> 'a -> bool
  val leta : 'a -> ('a -> 'b) -> 'b
end = struct

type 'a eq = {eq : 'a -> 'a -> bool};
val eq = #eq : 'a eq -> 'a -> 'a -> bool;

fun eqa A_ a b = eq A_ a b;

fun leta s f = f s;

end; (*struct HOL*)

structure IsaOption : sig
  val option_case : 'a -> ('b -> 'a) -> 'b option -> 'a
end = struct

fun option_case f1 f2 (SOME a) = f2 a
  | option_case f1 f2 NONE = f1;

end; (*struct IsaOption*)

structure IsaList : sig
  val foldr : ('a -> 'b -> 'b) -> 'a list -> 'b -> 'b
  val filter : ('a -> bool) -> 'a list -> 'a list
  val member : 'a HOL.eq -> 'a list -> 'a -> bool
  val insert : 'a HOL.eq -> 'a -> 'a list -> 'a list
  val removeAll : 'a HOL.eq -> 'a -> 'a list -> 'a list
end = struct

fun foldr f [] a = a
  | foldr f (x :: xs) a = f x (foldr f xs a);

fun filter p [] = []
  | filter p (x :: xs) = (if p x then x :: filter p xs else filter p xs);

fun member A_ [] y = false
  | member A_ (x :: xs) y = HOL.eqa A_ x y orelse member A_ xs y;

fun insert A_ x xs = (if member A_ xs x then xs else x :: xs);

fun removeAll A_ x [] = []
  | removeAll A_ x (y :: xs) =
    (if HOL.eqa A_ x y then removeAll A_ x xs else y :: removeAll A_ x xs);

end; (*struct IsaList*)

structure Set : sig
  datatype 'a set = Set of 'a list | Coset of 'a list
  val insert : 'a HOL.eq -> 'a -> 'a set -> 'a set
  val member : 'a HOL.eq -> 'a -> 'a set -> bool
end = struct

datatype 'a set = Set of 'a list | Coset of 'a list;

fun insert A_ x (Coset xs) = Coset (IsaList.removeAll A_ x xs)
  | insert A_ x (Set xs) = Set (IsaList.insert A_ x xs);

fun member A_ x (Coset xs) = not (IsaList.member A_ xs x)
  | member A_ x (Set xs) = IsaList.member A_ xs x;

end; (*struct Set*)

structure More_Set : sig
  val remove : 'a HOL.eq -> 'a -> 'a Set.set -> 'a Set.set
end = struct

fun remove A_ x (Set.Coset xs) = Set.Coset (IsaList.insert A_ x xs)
  | remove A_ x (Set.Set xs) = Set.Set (IsaList.removeAll A_ x xs);

end; (*struct More_Set*)

structure Executable_Set : sig
  val empty : 'a Set.set
  val union : 'a HOL.eq -> 'a Set.set -> 'a Set.set -> 'a Set.set
  val subtract : 'a HOL.eq -> 'a Set.set -> 'a Set.set -> 'a Set.set
end = struct

val empty : 'a Set.set = Set.Set [];

fun union A_ (Set.Coset xs) a =
  Set.Coset (IsaList.filter (fn x => not (Set.member A_ x a)) xs)
  | union A_ (Set.Set xs) a = IsaList.foldr (Set.insert A_) xs a;

fun subtract A_ (Set.Coset xs) a =
  Set.Set (IsaList.filter (fn x => Set.member A_ x a) xs)
  | subtract A_ (Set.Set xs) a = IsaList.foldr (More_Set.remove A_) xs a;

end; (*struct Executable_Set*)

structure IsaString : sig
  val eq_literal : string HOL.eq
end = struct

val eq_literal = {eq = (fn a => fn b => ((a : string) = b))} : string HOL.eq;

end; (*struct IsaString*)

structure CR : sig
(*  datatype filling = Fill*)
  type filling
  val Fill : filling

  datatype ground = Nt
  datatype tipi = Base of ground * filling | Funk of tipi * tipi * filling |
    Unit of filling | Sumtp of tipi * tipi * filling
  datatype lExp = LVar of string * filling |
    LLam of string * tipi * lExp * filling | LApp of lExp * lExp * filling |
    Unity of filling | Fp of string * tipi * lExp * filling |
    InL of tipi * lExp * filling | InR of tipi * lExp * filling |
    Caseof of lExp * string * lExp * string * lExp * filling | Zero of filling |
    Succ of lExp * filling | Pred of lExp * filling |
    Ifz of lExp * lExp * lExp * filling
(*  val fv : lExp -> string -> bool*)
  val fv : lExp -> string Set.set
  val lt : string -> tipi -> lExp -> lExp -> lExp
  val naiveSubst : lExp -> lExp -> string -> lExp
  val r2b :
    string -> tipi -> lExp -> lExp -> string -> string -> tipi -> lExp -> lExp
  val eq_filling : filling -> filling -> bool
  val eq_ground : ground -> ground -> bool
  val eq_tipi : tipi -> tipi -> bool
  val capchures : lExp -> lExp -> bool
  val contract : (string -> tipi option) -> string -> string -> tipi option
  val extend :
    (string -> tipi option) -> string -> tipi -> string -> tipi option
  val extendContract :
    (string -> tipi option) -> string -> tipi -> string -> tipi option
  val tipi_case :
    (ground -> filling -> 'a) ->
      (tipi -> tipi -> filling -> 'a) ->
        (filling -> 'a) -> (tipi -> tipi -> filling -> 'a) -> tipi -> 'a
  val typeinfer : lExp -> (string -> tipi option) -> tipi option
  val r2a :
    lExp ->
      lExp ->
        filling ->
          string -> string -> tipi -> lExp -> (string -> tipi option) -> lExp
  val r2 :
    lExp -> string -> string -> tipi -> lExp -> (string -> tipi option) -> lExp
end = struct

(*datatype filling = Fill;*)
type filling = string list;
val Fill = [" "," "]

datatype ground = Nt;

datatype tipi = Base of ground * filling | Funk of tipi * tipi * filling |
  Unit of filling | Sumtp of tipi * tipi * filling;

datatype lExp = LVar of string * filling |
  LLam of string * tipi * lExp * filling | LApp of lExp * lExp * filling |
  Unity of filling | Fp of string * tipi * lExp * filling |
  InL of tipi * lExp * filling | InR of tipi * lExp * filling |
  Caseof of lExp * string * lExp * string * lExp * filling | Zero of filling |
  Succ of lExp * filling | Pred of lExp * filling |
  Ifz of lExp * lExp * lExp * filling;

fun fv (LVar (i, fill)) = Set.insert IsaString.eq_literal i Executable_Set.empty
  | fv (LLam (i, t, e, fill)) =
    Executable_Set.subtract IsaString.eq_literal
      (Set.insert IsaString.eq_literal i Executable_Set.empty) (fv e)
  | fv (LApp (e1, e2, fill)) =
    Executable_Set.union IsaString.eq_literal (fv e1) (fv e2)
  | fv (Unity fill) = Executable_Set.empty
  | fv (Fp (i, t, e, fill)) =
    Executable_Set.subtract IsaString.eq_literal
      (Set.insert IsaString.eq_literal i Executable_Set.empty) (fv e)
  | fv (InL (t, m, fill)) = fv m
  | fv (InR (t, m, fill)) = fv m
  | fv (Caseof (m, x, l, y, n, fill)) =
    Executable_Set.union IsaString.eq_literal
      (Executable_Set.union IsaString.eq_literal (fv l)
        (Executable_Set.subtract IsaString.eq_literal
          (Set.insert IsaString.eq_literal x Executable_Set.empty) (fv m)))
      (Executable_Set.subtract IsaString.eq_literal
        (Set.insert IsaString.eq_literal y Executable_Set.empty) (fv n))
  | fv (Zero fill) = Executable_Set.empty
  | fv (Succ (m, fill)) = fv m
  | fv (Pred (m, fill)) = fv m
  | fv (Ifz (l, m, n, fill)) =
    Executable_Set.union IsaString.eq_literal
      (Executable_Set.union IsaString.eq_literal (fv l) (fv m)) (fv n);

fun lt x t n m = LApp (LLam (x, t, m, Fill), n, Fill);

fun naiveSubst (LVar (i, fill)) m x =
  (if ((x : string) = i) then m else LVar (i, fill))
  | naiveSubst (LLam (i, t, e, fill)) m x =
    (if ((x : string) = i) then LLam (i, t, e, fill)
      else LLam (i, t, naiveSubst e m x, fill))
  | naiveSubst (LApp (e1, e2, fill)) m x =
    LApp (naiveSubst e1 m x, naiveSubst e2 m x, fill)
  | naiveSubst (Unity fill) m x = Unity fill
  | naiveSubst (Fp (i, t, e, fill)) m x =
    (if ((x : string) = i) then Fp (i, t, e, fill)
      else Fp (i, t, naiveSubst e m x, fill))
  | naiveSubst (InL (t, n, fill)) m x = InL (t, naiveSubst n m x, fill)
  | naiveSubst (InR (t, n, fill)) m x = InR (t, naiveSubst n m x, fill)
  | naiveSubst (Caseof (m, x, l, y, n, fill)) k i =
    (if ((i : string) = x) andalso ((i : string) = y)
      then Caseof (m, x, naiveSubst l k i, y, n, fill)
      else (if ((i : string) = x) andalso not ((i : string) = y)
             then Caseof (m, x, naiveSubst l k i, y, naiveSubst n k i, fill)
             else (if not ((i : string) = x) andalso ((i : string) = y)
                    then Caseof
                           (naiveSubst m k i, x, naiveSubst l k i, y, n, fill)
                    else Caseof
                           (naiveSubst m k i, x, naiveSubst l k i, y,
                             naiveSubst n k i, fill))))
  | naiveSubst (Zero fill) m x = Zero fill
  | naiveSubst (Succ (n, fill)) m x = Succ (naiveSubst n m x, fill)
  | naiveSubst (Pred (n, fill)) m x = Pred (naiveSubst n m x, fill)
  | naiveSubst (Ifz (l, na, n, fill)) m x =
    Ifz (naiveSubst l m x, naiveSubst na m x, naiveSubst n m x, fill);

fun r2b xa ta n m x y t l =
  lt xa (Sumtp (ta, t, Fill)) (InL (Sumtp (ta, t, Fill), m, Fill))
    (naiveSubst n (Caseof (LVar (x, Fill), x, LVar (xa, Fill), y, l, Fill)) xa);

fun eq_filling Fill1 Fill2 = true;

fun eq_ground Nt Nt = true;

fun eq_tipi (Base (grounda, fillinga)) (Base (ground, filling)) =
  eq_ground grounda ground andalso eq_filling fillinga filling
  | eq_tipi (Funk (tipi1a, tipi2a, fillinga)) (Funk (tipi1, tipi2, filling)) =
    eq_tipi tipi1a tipi1 andalso
      (eq_tipi tipi2a tipi2 andalso eq_filling fillinga filling)
  | eq_tipi (Unit fillinga) (Unit filling) = eq_filling fillinga filling
  | eq_tipi (Sumtp (tipi1a, tipi2a, fillinga)) (Sumtp (tipi1, tipi2, filling)) =
    eq_tipi tipi1a tipi1 andalso
      (eq_tipi tipi2a tipi2 andalso eq_filling fillinga filling)
  | eq_tipi (Base (ground, fillinga)) (Funk (tipi1, tipi2, filling)) = false
  | eq_tipi (Funk (tipi1, tipi2, fillinga)) (Base (ground, filling)) = false
  | eq_tipi (Base (ground, fillinga)) (Unit filling) = false
  | eq_tipi (Unit fillinga) (Base (ground, filling)) = false
  | eq_tipi (Base (ground, fillinga)) (Sumtp (tipi1, tipi2, filling)) = false
  | eq_tipi (Sumtp (tipi1, tipi2, fillinga)) (Base (ground, filling)) = false
  | eq_tipi (Funk (tipi1, tipi2, fillinga)) (Unit filling) = false
  | eq_tipi (Unit fillinga) (Funk (tipi1, tipi2, filling)) = false
  | eq_tipi (Funk (tipi1a, tipi2a, fillinga)) (Sumtp (tipi1, tipi2, filling)) =
    false
  | eq_tipi (Sumtp (tipi1a, tipi2a, fillinga)) (Funk (tipi1, tipi2, filling)) =
    false
  | eq_tipi (Unit fillinga) (Sumtp (tipi1, tipi2, filling)) = false
  | eq_tipi (Sumtp (tipi1, tipi2, fillinga)) (Unit filling) = false;

fun capchures (LVar (i, f)) m = false
  | capchures (LLam (i, t, ma, f)) m =
    Set.member IsaString.eq_literal i (fv m) orelse capchures ma m
  | capchures (LApp (m1, m2, f)) m = capchures m1 m orelse capchures m2 m
  | capchures (Unity f) m = false
  | capchures (Fp (i, t, ma, f)) m =
    Set.member IsaString.eq_literal i (fv m) orelse capchures ma m
  | capchures (InL (t, ma, f)) m = capchures ma m
  | capchures (InR (t, ma, f)) m = capchures ma m
  | capchures (Caseof (ma, x, l, y, n, f)) m =
    capchures l m orelse
      (Set.member IsaString.eq_literal x (fv m) orelse
        (capchures ma m orelse
          (Set.member IsaString.eq_literal y (fv m) orelse capchures n m)))
  | capchures (Zero f) m = false
  | capchures (Succ (ma, f)) m = capchures ma m
  | capchures (Pred (ma, f)) m = capchures ma m
  | capchures (Ifz (l, ma, n, f)) m =
    capchures l m orelse (capchures ma m orelse capchures n m);

fun contract gamma x =
  (fn v => (if not ((v : string) = x) then gamma v else NONE));

fun extend gamma x t =
  (fn v => (if ((v : string) = x) then SOME t else gamma v));

fun extendContract gamma x t = extend (contract gamma x) x t;

fun tipi_case f1 f2 f3 f4 (Sumtp (tipi1, tipi2, filling)) =
  f4 tipi1 tipi2 filling
  | tipi_case f1 f2 f3 f4 (Unit filling) = f3 filling
  | tipi_case f1 f2 f3 f4 (Funk (tipi1, tipi2, filling)) =
    f2 tipi1 tipi2 filling
  | tipi_case f1 f2 f3 f4 (Base (ground, filling)) = f1 ground filling;

fun typeinfer (LVar (i, f)) g = g i
  | typeinfer (LLam (x, t, m, f)) g =
    (case typeinfer m (extendContract g x t) of NONE => NONE
      | SOME ta => SOME (Funk (t, ta, f)))
  | typeinfer (LApp (e1, e2, f)) g =
    (case typeinfer e1 g of NONE => NONE | SOME (Base (_, _)) => NONE
      | SOME (Funk (t, ta, _)) =>
        (case typeinfer e2 g of NONE => NONE
          | SOME tb => (if eq_tipi tb t then SOME ta else NONE))
      | SOME (Unit _) => NONE | SOME (Sumtp (_, _, _)) => NONE)
  | typeinfer (Unity f) g = SOME (Unit Fill)
  | typeinfer (Fp (i, t, e, f)) g =
    (case typeinfer e (extendContract g i t) of NONE => NONE
      | SOME ta => (if eq_tipi t ta then SOME t else NONE))
  | typeinfer (InL (t, m, f)) g =
    (case t of Base (_, _) => NONE | Funk (_, _, _) => NONE | Unit _ => NONE
      | Sumtp (ta, _, _) =>
        (case typeinfer m g of NONE => NONE
          | SOME tb => (if eq_tipi tb ta then SOME t else NONE)))
  | typeinfer (InR (t, m, f)) g =
    (case t of Base (_, _) => NONE | Funk (_, _, _) => NONE | Unit _ => NONE
      | Sumtp (_, ta, _) =>
        (case typeinfer m g of NONE => NONE
          | SOME tb => (if eq_tipi tb ta then SOME t else NONE)))
  | typeinfer (Caseof (m, x, l, y, n, f)) g =
    (case typeinfer l g of NONE => NONE | SOME (Base (_, _)) => NONE
      | SOME (Funk (_, _, _)) => NONE | SOME (Unit _) => NONE
      | SOME (Sumtp (t, ta, _)) =>
        (case typeinfer m (extendContract g x t) of NONE => NONE
          | SOME s =>
            (case typeinfer n (extendContract g y ta) of NONE => NONE
              | SOME sa => (if eq_tipi s sa then SOME s else NONE))))
  | typeinfer (Zero f) g = SOME (Base (Nt, Fill))
  | typeinfer (Succ (m, f)) g =
    (case typeinfer m g of NONE => NONE
      | SOME t => (if eq_tipi t (Base (Nt, Fill)) then SOME t else NONE))
  | typeinfer (Pred (m, f)) g =
    (case typeinfer m g of NONE => NONE
      | SOME t => (if eq_tipi t (Base (Nt, Fill)) then SOME t else NONE))
  | typeinfer (Ifz (l, m, n, f)) g =
    (case typeinfer l g of NONE => NONE
      | SOME t =>
        (if not (eq_tipi t (Base (Nt, Fill))) then NONE
          else (case typeinfer m g of NONE => NONE
                 | SOME ta =>
                   (case typeinfer n g of NONE => NONE
                     | SOME taa =>
                       (if eq_tipi ta taa then SOME ta else NONE)))));

fun r2a (LVar (i, f)) m f2 uu uua uub uuc uud = LApp (LVar (i, f), m, f2)
  | r2a (LLam (xa, ta, n, f)) m f2 x y t l g =
let
val _ = if (not (!PCFP1_Syntax.debugging))
	then ()
	else
	    let
		val _ = print "...side-conditions check:\n"
		val c1 = typeinfer l (extendContract (extendContract g xa ta) y t)
		val _ = if (Option.isSome c1)
			then let
				val _ = print "\ttypecheck succeeds\n"
				val tb = Option.valOf c1
				val c8 = eq_tipi ta tb
				val _ = if c8
					then print "\ttype-equality test positive\n"
					else print "\ttype-equality test negative!\n"
			    in ()
			    end
			else print "\ttypecheck fails!\n"
		val c2 = (capchures n
                       (Caseof
                         (LVar (x, Fill), x, LVar (xa, Fill), y, l,
                           Fill)))
		val _ = if c2
			then print "\tcapture test 1 positive!\n"
			else print "\tcapture test 1 negative\n"
		val c3 = (capchures n m)
		val _ = if c3
			then print "\tcapture test 2 positive!\n"
			else print "\tcapture test 2 negative\n"
		val c4 = (capchures l m)
		val _ = if c4
			then print "\tcapture test 3 positive!\n"
			else print "\tcapture test 3 negative\n"
		val c5 = (Set.member IsaString.eq_literal x (fv m))
		val _ = if c5
			then print "\tmembership test 1 positive!\n"
			else print "\tmembership test 1 negative\n"
		val c6 = (Set.member IsaString.eq_literal y (fv m))
		val _ = if c6
			then print "\tmembership test 2 positive!\n"
			else print "\tmembership test 2 negative\n"
		val c7 = (Set.member IsaString.eq_literal xa (fv l))
		val _ = if c7
			then print "\tmembership test 3 positive!\n"
			else print "\tmembership test 3 negative\n"
		val _ = print "...end\n"
	    in ()
	    end
in
    (case typeinfer l (extendContract (extendContract g xa ta) y t)
      of NONE => LApp (LLam (xa, ta, n, f), m, f2)
      | SOME tb =>
        (if not (not (capchures n
                       (Caseof
                         (LVar (x, Fill), x, LVar (xa, Fill), y, l,
                           Fill))) andalso
                  (not (capchures n m) andalso
                    (not (capchures l m) andalso
                      (not (Set.member IsaString.eq_literal x (fv m)) andalso
                        (not (Set.member IsaString.eq_literal y (fv m)) andalso
                          (not (Set.member IsaString.eq_literal xa (fv l)) andalso
                            eq_tipi ta tb))))))
          then LApp (LLam (xa, ta, n, f), m, f2) else r2b xa ta n m x y t l))
end
  | r2a (LApp (e1, e2, f)) m f2 uu uua uub uuc uud =
    LApp (LApp (e1, e2, f), m, f2)
  | r2a (Unity f) m f2 uu uua uub uuc uud = LApp (Unity f, m, f2)
  | r2a (Fp (i, t, e, f)) m f2 uu uua uub uuc uud =
    LApp (Fp (i, t, e, f), m, f2)
  | r2a (InL (t, ma, f)) m f2 uu uua uub uuc uud = LApp (InL (t, ma, f), m, f2)
  | r2a (InR (t, ma, f)) m f2 uu uua uub uuc uud = LApp (InR (t, ma, f), m, f2)
  | r2a (Caseof (ma, x, l, y, n, f)) m f2 uu uua uub uuc uud =
    LApp (Caseof (ma, x, l, y, n, f), m, f2)
  | r2a (Zero f) m f2 uu uua uub uuc uud = LApp (Zero f, m, f2)
  | r2a (Succ (ma, f)) m f2 uu uua uub uuc uud = LApp (Succ (ma, f), m, f2)
  | r2a (Pred (ma, f)) m f2 uu uua uub uuc uud = LApp (Pred (ma, f), m, f2)
  | r2a (Ifz (l, ma, n, f)) m f2 uu uua uub uuc uud =
    LApp (Ifz (l, ma, n, f), m, f2);

fun r2 (LVar (i, f)) uu uua uub uuc uud = LVar (i, f)
  | r2 (LLam (i, t, e, f)) uu uua uub uuc uud = LLam (i, t, e, f)
  | r2 (LApp (e1, e2, f)) x y t l g = r2a e1 e2 f x y t l g
  | r2 (Unity f) uu uua uub uuc uud = Unity f
  | r2 (Fp (i, t, e, f)) uu uua uub uuc uud = Fp (i, t, e, f)
  | r2 (InL (t, m, f)) uu uua uub uuc uud = InL (t, m, f)
  | r2 (InR (t, m, f)) uu uua uub uuc uud = InR (t, m, f)
  | r2 (Caseof (m, x, l, y, n, f)) uu uua uub uuc uud =
    Caseof (m, x, l, y, n, f)
  | r2 (Zero f) uu uua uub uuc uud = Zero f
  | r2 (Succ (m, f)) uu uua uub uuc uud = Succ (m, f)
  | r2 (Pred (m, f)) uu uua uub uuc uud = Pred (m, f)
  | r2 (Ifz (l, m, n, f)) uu uua uub uuc uud = Ifz (l, m, n, f);

end; (*struct CR*)
