(*
#$Id: pcfp1.lex 34 2011-07-21 17:52:28Z ns441 $
#
#Part of the "prefactor" experiment. Scanner definition.
#
#Nik Sultana
#www.cl.cam.ac.uk/~ns441
#
*)

structure T = Tokens
(*Position in file*)
type pos = int
type lineNo = int
type svalue = T.svalue
type ('a,'b) token = ('a,'b) T.token
type lexresult = (svalue,pos) token
type lexarg = string
type arg = lexarg
val col = ref 0;
(*Line pointer*)
val linep = ref 1;
val eolpos = ref 0;

val eof = fn fileName =>
  let
    val result = T.EOF (!linep,!col);
    val _ = linep := 0;
  in result end

%%
%header (functor PCF_LexF(structure Tokens: PCF_TOKENS));
%arg (fileName:string);

printable_char = .;
viewable_char = [.\n];
numeric = [0-9];
lower_alpha = [a-z];
upper_alpha = [A-Z];
alpha_numeric = ({lower_alpha}|{upper_alpha}|{numeric}|_);

ws = ([\ ]|[\t])*;
eol = ("\013\010"|"\010"|"\013");

%%

"nat" => (col:=yypos-(!eolpos); T.T_NAT(!linep,!col));
"->" => (col:=yypos-(!eolpos); T.T_FN(!linep,!col));
"+" => (col:=yypos-(!eolpos); T.T_SUM(!linep,!col));
"unit" => (col:=yypos-(!eolpos); T.T_UNIT(!linep,!col));

{ws} => (col:=yypos-(!eolpos); T.WHITESPACE(yytext,!linep,!col));

{eol} => (linep:=(!linep)+1;
		      eolpos:=yypos+size yytext;
          T.NEWLINE(!linep,!col));

"fn" => (col:=yypos-(!eolpos); T.E_FN(!linep,!col));
"=>" => (col:=yypos-(!eolpos); T.E_RARROW(!linep,!col));
"<=" => (col:=yypos-(!eolpos); T.E_LARROW(!linep,!col));
":" => (col:=yypos-(!eolpos); T.E_COLON(!linep,!col));
"unity" => (col:=yypos-(!eolpos); T.E_UNITY(!linep,!col));
"fix" => (col:=yypos-(!eolpos); T.E_FIX(!linep,!col));
"left" => (col:=yypos-(!eolpos); T.E_LEFT(!linep,!col));
"right" => (col:=yypos-(!eolpos); T.E_RIGHT(!linep,!col));
"case" => (col:=yypos-(!eolpos); T.E_CASE(!linep,!col));
"zero" => (col:=yypos-(!eolpos); T.E_ZERO(!linep,!col));
"succ" => (col:=yypos-(!eolpos); T.E_SUCC(!linep,!col));
"pred" => (col:=yypos-(!eolpos); T.E_PRED(!linep,!col));
"ifZero" => (col:=yypos-(!eolpos); T.E_IFZERO(!linep,!col));

"then" => (col:=yypos-(!eolpos); T.E_THEN(!linep,!col));
"else" => (col:=yypos-(!eolpos); T.E_ELSE(!linep,!col));

"(" => (col:=yypos-(!eolpos); T.LPAREN(!linep,!col));
")" => (col:=yypos-(!eolpos); T.RPAREN(!linep,!col));

"[" => (col:=yypos-(!eolpos); T.LBRACK(!linep,!col));
"]" => (col:=yypos-(!eolpos); T.RBRACK(!linep,!col));

{alpha_numeric}* => (col:=yypos-(!eolpos); T.E_IDENT(yytext,!linep,!col));
