open PCFP1_Syntax

local
  (*map option string to string*)
  fun optionstring fill = Option.getOpt (fill, "")
  fun set_outerfilling fill left_fill right_fill =
    if List.null fill then
      [optionstring left_fill, optionstring right_fill]
    else
      let
        val fill_hd = optionstring left_fill ^ hd fill
        val fill_tl = List.rev (tl fill)
	    in
	      if List.null fill_tl then
	        [fill_hd, optionstring right_fill]
	      else
	        fill_hd :: (List.rev ((hd fill_tl ^ optionstring right_fill) ::
	          tl fill_tl)) end
in
  fun set_outerfilling_of_type (Type (ty, fill)) left_fill right_fill =
    Type (ty, set_outerfilling fill left_fill right_fill)
  fun set_outerfilling_of_term (Term (t, fill)) left_fill right_fill =
    Term (t, set_outerfilling fill left_fill right_fill)
end

%%
%name PCF
%term T_NAT
| T_FN
| T_SUM
| T_UNIT
| WHITESPACE of string
| NEWLINE
| E_FN
| E_RARROW
| E_LARROW
| E_COLON
| E_UNITY
| E_FIX
| E_LEFT
| E_RIGHT
| E_CASE
| E_ZERO
| E_SUCC
| E_PRED
| E_IFZERO
| LPAREN
| RPAREN
| E_IDENT of string
| EOF
| APPLY
| E_THEN
| E_ELSE
| LBRACK
| RBRACK

%nonterm expression of expression
| typ of typ
| variable of string pre_trm
| filling of string
| ground_typ of pre_typ
| atom of string pre_trm
| trm of string trm

%pos int
%eop EOF
%noshift EOF
%arg (fileName) : string

%nonassoc E_FN E_UNITY E_FIX E_LEFT E_RIGHT E_ZERO E_SUCC E_PRED E_IFZERO
%nonassoc T_UNIT
%nonassoc T_NAT
%nonassoc E_RARROW E_LARROW E_COLON

%left NEWLINE
%left WHITESPACE
%nonassoc E_CASE
%right T_SUM
%right T_FN

%left APPLY

%start expression
%pure
%verbose
%%

(*
#$Id: pcfp1.yacc 32 2011-07-20 23:47:15Z ns441 $
#
#Part of the "prefactor" experiment. Parser definition.
#
#Nik Sultana
#www.cl.cam.ac.uk/~ns441
#
*)

expression : trm (( TermExp trm ))
           | typ (( TypeExp typ))

filling : WHITESPACE (( WHITESPACE ))
        | NEWLINE   (( "\n" ))

ground_typ : T_NAT       (( T_Nat ))
           | T_UNIT      (( T_Unit ))

typ  : typ T_FN typ      (( Type (T_Fun (typ1, typ2), ["",""]) ))
     | typ T_SUM typ     (( Type (T_Sum (typ1, typ2), ["",""]) ))
     | ground_typ        (( Type (ground_typ, ["",""]) ))
     | filling typ       %prec WHITESPACE (( set_outerfilling_of_type typ (SOME filling) NONE ))
     | typ filling       %prec WHITESPACE (( set_outerfilling_of_type typ NONE (SOME filling) ))
     | LPAREN typ RPAREN (( set_outerfilling_of_type typ (SOME "(") (SOME ")") ))

variable : E_IDENT (( E_Var E_IDENT ))

atom : variable  (( variable ))
     | E_UNITY   (( E_Un ))
     | E_ZERO    (( E_Zero  ))

trm : atom              (( Term (atom, ["",""]) ))
    | filling trm       %prec WHITESPACE (( set_outerfilling_of_term trm
                                             (SOME filling) NONE ))
    | trm filling       %prec WHITESPACE (( set_outerfilling_of_term trm
                                             NONE (SOME filling) ))
    | LPAREN trm RPAREN (( set_outerfilling_of_term trm
                                             (SOME "(") (SOME ")") ))
    | trm WHITESPACE trm %prec APPLY   (( Term (E_App (trm1, trm2),
                                            ["",WHITESPACE,""])  ))
    | E_FN WHITESPACE E_IDENT WHITESPACE E_COLON WHITESPACE typ WHITESPACE E_RARROW filling trm ((
        Term (E_Lam (E_IDENT, typ, trm), ["", WHITESPACE1, WHITESPACE2,
          WHITESPACE3, WHITESPACE4, filling, ""]) ))
    | E_FIX WHITESPACE E_IDENT WHITESPACE E_COLON WHITESPACE typ WHITESPACE E_RARROW filling trm ((
        Term (E_Fix (E_IDENT, typ, trm), ["", WHITESPACE1, WHITESPACE2,
          WHITESPACE3, WHITESPACE4, filling, ""]) ))
    | E_SUCC WHITESPACE trm ((
        Term (E_Succ trm, ["", WHITESPACE, ""]) ))
    | E_PRED WHITESPACE trm ((
        Term (E_Pred trm, ["", WHITESPACE, ""]) ))
    | E_IFZERO WHITESPACE trm filling E_THEN WHITESPACE trm filling E_ELSE WHITESPACE trm ((
        Term (E_IfZero (trm1, trm2, trm3), ["", WHITESPACE1, filling1,
          WHITESPACE2, filling2, WHITESPACE3, ""]) ))
    | E_LEFT WHITESPACE typ WHITESPACE trm ((
        Term (E_InL (typ, trm), ["", WHITESPACE1, WHITESPACE2, ""]) ))
    | E_RIGHT WHITESPACE typ WHITESPACE trm ((
        Term (E_InR (typ, trm), ["", WHITESPACE1, WHITESPACE2, ""]) ))
    | E_CASE WHITESPACE LBRACK WHITESPACE trm WHITESPACE E_LARROW WHITESPACE E_IDENT WHITESPACE RBRACK WHITESPACE trm WHITESPACE LBRACK WHITESPACE E_IDENT WHITESPACE E_RARROW WHITESPACE trm WHITESPACE RBRACK ((
        Term (E_Case (trm1, E_IDENT1, trm2, E_IDENT2, trm3), ["",WHITESPACE1,
          WHITESPACE2, WHITESPACE3, WHITESPACE4, WHITESPACE5, WHITESPACE6,
          WHITESPACE7, WHITESPACE8, WHITESPACE9, WHITESPACE10, WHITESPACE11,
          ""]) ))
